// Mollification routines taken from Diego Murio
 
/**************************************************************************
*                                                                         *
*          Automatic Numerical Differentiation Procedure                  *
*     Read the main() to find out how to use these procedures.            *
*                                                                         *
*          Copyright 1997-Present Diego A. Murio                          *
*                                                                         *
***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "mollify.H"

int itmin;
double *wt;


/* use golden search to find the mollification radius by GCV method and then */
/* filter the data with this radius */
void gcvmol(int kt, double z[SIZE])
{
   double gcv(int kt, double delta, double *z);

   int i, niter, maxiter;
   double zt[SIZE];
   double tol, dmin, dmax, cc, gr, delta, gold;
   double x0, x1, x2, x3, afirst, f1, f2;

   for (i = 0; i <= kt; ++i)
      zt[i] = z[i];

   tol = 0.0001; dmin = 0.001; dmax = 0.16;
   maxiter = 100; gr =  0.5 * (sqrt(5) - 1); cc = 1 - gr;

   delta = 0.04; niter = 0; 
   x0 = dmin; x3 = dmax;
   if (fabs(dmax - delta) > fabs(delta - dmin)){
      x1 = delta;
      x2 = delta + cc * (dmax - delta);
   }
   else{
      x2 = delta;
      x1 = delta - cc * (delta - dmin);
   }
   f1 = gcv(kt, x1, z);
   for (i = 0; i <= kt; ++i)
      z[i] = zt[i];
   f2 = gcv(kt, x2, z);
   afirst = tol * (fabs(x1) + fabs(x2));

   while (fabs(x3 - x0) > afirst && niter < maxiter){
      if ( f2 < f1){
         x0 = x1; x1 = x2; x2 = gr * x1 + cc * x3;
         f1 = f2;
         for (i = 0; i <= kt; ++i)
             z[i] = zt[i];
         f2 = gcv(kt, x2, z);
      }
      else{
         x3 = x2; x2 = x1; x1 = gr * x2 + cc * x0;
         f2 = f1;
         for (i = 0; i <= kt; ++i)
             z[i] = zt[i];
         f1 = gcv(kt, x1, z);
      }
      afirst = tol * (fabs(x1) + fabs(x2));
      niter += 1;
   }
   if (f1 < f2)
      delta = x1;
   else
      delta = x2;
   if (niter == maxiter) delta = dmin;
   
   for (i = 0; i <= kt; ++i)
      z[i] = zt[i];

   gold = gcv(kt, delta, z);
    
}

/* calculation of the GCV function in order to choose a delta which */
/* minimizes this function */  
double gcv(int kt, double delta, double z[SIZE])
{
   void weight(int kt, double delta);
   void mol(int kt, double *z);

   int i;
   double zt[SIZE], sum, sumgcv;

   for (i = 0; i <= kt; ++i)
      zt[i] = z[i];

   weight(kt, delta);
   mol(kt, z);

   sum = 0;
   for (i = 0; i <= kt; ++i)
      sum += (zt[i] - z[i]) * (zt[i] - z[i]);

   sumgcv = (double) itmin;
   for (i = 1; i <= itmin; ++i)
      sumgcv += (2.0 * i - 1.0) * pow(wt[itmin-i+1], 2);

   if (wt != 0)
      free(wt);
   return sum/sumgcv;
}

/* after the user inputs delta, this procedure filters the input data */
void deltaMol(int kt, double delta, double z[SIZE])
{
   void weight(int kt, double delta);
   void mol(int kt, double z[SIZE]);

   weight(kt, delta);
   mol(kt, z);

}

/* after computation of mollification weights, we compute the mollified data */
void mol(int kt, double z[SIZE])
{
   int i,j;
   double s1, s2, p1, p2, c1, c2, um[SIZE], temp[SIZE];

   /* calculation of extension constant near 1 */
   s1 = 0; s2 = 0;
   for (i = kt - itmin + 1; i <= kt; ++i){
      p1 = 0; p2 = 0;
      for (j = kt - i + itmin; j < 2 * itmin; ++j)
         p1 += wt[j];
      for (j = i - itmin; j < kt; ++j)
         p2 += wt[j-i+itmin] * z[j];
      s1 += (z[i] - p2) * p1;
      s2 += p1 * p1;
   }
   c2 = s1 / s2;

   /* calculation of extension constant near 0 */
   for (i = 0; i <= kt; ++i)
       temp[i] = z[kt-i];
   s1 = 0; s2 = 0;
   for (i = kt - itmin + 1; i <= kt; ++i){
      p1 = 0; p2 = 0;
      for (j = kt - i + itmin; j < 2 * itmin; ++j)
         p1 += wt[j];
      for (j = i - itmin; j < kt; ++j)
         p2 += wt[j-i+itmin] * temp[j];
      s1 += (temp[i] - p2) * p1;
      s2 += p1 * p1;
   }
   c1 = s1 / s2;

   /* mollification for [0, 3delta) */
   for (i = 0; i < itmin; ++i){
      s1 = 0; s2 = 0;
      for (j = 0; j < itmin - i; ++j)
         s1 += wt[j];
      for (j = itmin - i; j <= 2 * itmin; ++j)
         s2 += wt[j] * z[i-itmin+j];
      um[i] = s2 + s1 * c1;
   }

   /* mollification for [3delta, 1-3delta] */
   for (i = itmin; i <= kt - itmin; ++i){
      s1 = 0;
      for (j = 0; j <= 2 * itmin; ++j)
         s1 += wt[j] * z[i-itmin+j];
      um[i] = s1;
   }
   
   /* mollification for (1-3delta, 1] */
   for (i = kt - itmin + 1; i <= kt; ++i){
      s1 = 0; s2 = 0;
      for (j = kt - i + itmin + 1; j <= 2 * itmin; ++j)
         s1 += wt[j];
      for (j = 0; j <= kt - i + itmin; ++j)
         s2 += wt[j] * z[i-itmin+j];
      um[i] = s2 + s1 * c2;
   }

   /* new z --- mollified z */
   for (i = 0; i <= kt; ++i)
      z[i] = um[i];

}

/* computation of mollification weights */
void weight(int kt, double delta)
{
   double mdnord(double);
   int i, iwtmax;
   double y1, y2, dt;

   dt = 1 / (double) kt;
   itmin = floor(3 * delta / dt) + 1;
   iwtmax = 2 * itmin;
    wt = (double*)calloc(iwtmax+1, sizeof(double));
   //wt = new double[iwtmax+1];

   for (i = 0; i <= iwtmax; ++i){
      y1 = dt * sqrt(2) / delta;
      y2 = y1 * (i + 0.5 - itmin);
      y1 *= (i - 0.5 - itmin);
      wt[i] = mdnord(y2) - mdnord(y1);
   }

}

/* 2/sqrt(pi) * integral of e^(-x^2) over (0, x) */
double myerf(double x)
{
   double pi, Ap, f, q, s;
   int i, itest;

   pi = acos(-1);
   Ap = 1 / sqrt(pi);

   if (x >= 5.55) return 1;

   if (x > 3.55){
      f = 1;
      s = 1 / x;
      for (i = 1; i < 87; ++i){
         f *= 2 * i - 1;
         q = f * pow(-1, (double) i);
         q /= pow(2, (double) i);
         q /= pow(x, (double) (2 * i + 1));
         if (fabs(q) <= pow(10,-35)) break;
         s += q;
         itest = floor(pow(x, 2) - 0.5);
         if (i > itest) break;
      }
      s *= Ap * exp(-pow(x, 2));
      return (1 - s);
   }

   if (x > 0){
      f = 1;
      s = x;
      for (i = 1; i < 101; ++i){
         f *= i;
         q = pow(-1, (double) i);
         q *= pow(x, (double) (2 * i + 1));
         q /= f * (2 * i + 1);
         if (fabs(q) <= pow(10, -33)) break;
         s += q;
      }
      s *= 2 * Ap;
      return s;
   }

   return 0;

}
   
/* 1/sqrt(pi) * integral of e^(-x^2) over (-infty, y/sqrt(2)) */
double mdnord(double y)
{
   double myerf(double);
   double p;

   p = -y / sqrt(2);

   if (p > 13.2) return 0;
   if (p < -13.2) return 1;
   if (p >= -13.2 && p < 0)
      return 1 - 0.5 * (1 - myerf(-p));

   return 0.5 * (1 - myerf(p));
}


