/* Code from Jeff Uhlmann
Hi Joseph,

Below is code for computing the minimum-determinant
value for the generalized variant of CI that allows
for the optimal inclusion of known independent error
information.

A description of the generalization and its use was
provided in the postscript file that I sent previously.

- Jeff

P.S. It's kind of nasty code because I pulled it out
of my other code and tried to make it self-contained
for someone who didn't know how to deal with matrices
as arguments in C.  It worked for him, but you might
want to change the interfaces to accommodate the way
you allocated matrices.
  
To make things simple for people who program in C but
don't know good tricks for dealing with 2D arrays, this
code assumes that arrays are allocated as MaxDims x MaxDims
even if the actual matrix sizes are less.  This means
that any matrix can be passed as an argument as long as
parameters are passed to say what its real dimensions are.
*/


#include "ciomega.H"

#define EPSILON 0.0000001
#define SQRT_EPSILON 0.0001
#define AlmostInfinite 9999999.99
#define SMALL 1e-12
#define TINY 1e-15
#define NUMIT 25

int ludcmp(MMatrix &a, int *indx, double *d, int n)
{
   int i=0, imax=0, j=0, k=0;
   double big, dum, sum, temp;
   double vv[MaxDims];

   *d = (1.0);
   for (i=0; i<n; i++) {
      big = (0.0);
      for (j=0; j<n; j++)
         if ((temp = (fabs)(a(i,j))) > big) big = temp;
       if ((fabs) (big) <= (double) TINY) return (-1);
       vv[i] = (1.0) / big;
   }
   for (j=0; j<n; j++) {
      for (i=0; i<j; i++) {
         sum = a(i,j);
         for (k=0; k<i; k++) sum -= a(i,k)*a(k,j);
         a(i,j) = sum;
      }
      big = (0.0);    
      for (i=j; i<n; i++) {
         sum = a(i,j);
         for (k=0; k<j; k++) sum -= a(i,k)*a(k,j);
         a(i,j) = sum;
         if ((dum = vv[i] * (fabs) (sum)) >= big) {
            big = dum;
            imax = i;
         }
      }
      if (j != imax) {
         for (k=0; k<n; k++) {
            dum = a(imax,k);
            a(imax,k) = a(j,k);
            a(j,k) = dum;
         }
         *d = -(*d);
         vv[imax] = vv[j];
      }
      indx[j] = imax;
      if ((fabs) (a(j,j)) <= (double) TINY) a(j,j) = TINY;
      if (j != (n-1)) {
         dum = (1.0) / (a(j,j));
          for (i=j+1; i<n; i++) a(i,j) *= dum;
      }
   } 
   return(1);
}

void lubksb(MMatrix &a, int indx[MaxDims], double b[MaxDims], int n)
{
   int i, ii = (-1), ip, j;
   double sum;

   for (i=0; i<n; i++) {
      ip = indx[i];
      sum = b[ip];
      b[ip] = b[i];
      if (ii >= 0) for (j=ii; j<=(i-1); j++) sum -= a(i,j)*b[j];
      else if (sum) ii = i;
      b[i] = sum;
   }
   for (i=n-1; i>=0; i--) {
      sum = b[i];
      for (j=i+1; j<n; j++)
        sum -= a(i,j)*b[j];
      b[i] = sum/a(i,i);
   }
}


int Invert(MMatrix &m, MMatrix &mi, int n)
{
   double col[MaxDims];
   MMatrix dum(n,n);
   int ind[MaxDims];
   double d, scale, iscale;
   int i, j, res;

   scale = SMALL;
 
   for (j=0; j<n; j++)
     for (i=0; i<n; i++)
       if (m(j,i) > scale) scale = m(j,i);
 
   if ((scale>1) || (scale==SMALL)) scale = iscale = 1.0;
   else iscale = 1.0 / scale;
   for (j=0; j<n; j++)
      for (i=0; i<n; i++) dum(j,i) = m(j,i)*iscale;

   res = ludcmp(dum, ind, &d, n);
   if (res==-1) return(res);
   for (i=0; i<n; i++) d *= dum(i,i);
 
   if (fabs(d) <= TINY) return(-1);
   for (j=0; j<n; j++) {
      for (i=0; i<n; i++) col[i] = 0.0;
      col[j] = 1.0;
      lubksb(dum, ind, col, n);
      for (i=0; i<n; i++) mi(i,j) = col[i];
   }
   for (j=0; j<n; j++)
      for (i=0; i<n; i++) mi(j,i) *= iscale;
 
  return(0);
}

double LogDet(MMatrix &A, int n)
{
   MMatrix B(n,n);
   int i, j, k;
   long double sum;

   B(0,0) = sqrt(A(0,0));
   for (i=1; i<n; i++) {
      for (j=0; j<i; j++) {
         sum = 0;
         for (k=0; k<j; k++) sum += B(i,k) * B(j,k);
         B(i,j) = (A(i,j) - sum)/B(j,j);
      }
      sum = B(i,0) * B(i,0);
      for (k=1; k<i; k++) sum += B(i,k) * B(i,k);
      B(i,i) = A(i,i) - sum;
      if (B(i,i) <= 0.0) return(-AlmostInfinite);
      B(i,i) = sqrt((double)B(i,i));
   }
   
   sum = 1.0;
   for (i=0; i<n; i++) sum *= (double)B(i,i); 
   return((double)1.0/(sum));
}

double LogDetOfUpdate(double w, MMatrix &A1, MMatrix &A2, int n,
                      MMatrix &B1, MMatrix &B2, int m, MMatrix &H) 
{
   MMatrix A(n,n);
   MMatrix S(m,m);
   MMatrix Sinv(m,m);
   MMatrix AH(n,m);

   int i, j, k;
   double sum, weight;

   if (w < SQRT_EPSILON) {
      // printf("w -> 0.0\n");
      if (m < n) return(AlmostInfinite);
      for (i=0; i<m; i++) { 
         for (j=0; j<i; j++) 
            S(i,j) = S(j,i) = B1(i,j) + B2(i,j);
         S(i,i) = B1(i,i) + B2(i,i);
      }
      return(LogDet(S,m));
   }
   if ((1-w) < SQRT_EPSILON) {
      // printf("w -> INF\n");
      for (i=0; i<n; i++) { 
         for (j=0; j<i; j++) 
            A(i,j) = A(j,i) = A1(i,j) + A2(i,j);
         A(i,i) = A1(i,i) + A2(i,i);
      }
      return(LogDet(A,n));
   }
   // joe's way
   MMatrix tmp1(n,n);
   MMatrix tmp2(m,m);
   MMatrix A12 = A1 + A2;
   MMatrix B12 = B1 + B2;
   Invert(A12, tmp1, n);
   Invert(B1, tmp2, m);
   MMatrix tmplog = w*tmp1 + (1.0 - w)*(transp(H) * tmp2 * H);
   return(LogDet(tmplog, n)); 

/*
   weight = 1.0/w;
   for (i=0; i<n; i++) {
      for (j=0; j<i; j++) 
         A(i,j) = A(j,i) = weight*A1(i,j) + A2(i,j);
      A(i,i) = weight*A1(i,i) + A2(i,i);
   }
   for (i=0; i<n; i++) {
      for (j=0; j<m; j++) {
         sum = A(i,0)*H(j,0);
         for (k=1; k<n; k++) sum += A(i,k)*H(j,k);
         AH(i,j) = sum;
      }
   }
   weight = 1.0/(1-w);
   for (i=0; i<m; i++) {
      for (j=0; j<i; j++) {
         sum = H(i,0)*AH(0,j);
         for (k=1; k<n; k++) sum += H(i,k)*AH(k,j);
         S(i,j) = S(j,i) = sum + weight*B1(i,j) + B2(i,j);
      }
      sum = H(i,0)*AH(0,i);
      for (k=1; k<n; k++) sum += H(i,k)*AH(k,i);
      S(i,i) = sum + weight*B1(i,i) + B2(i,i); 
   }         
   Invert(S,Sinv,m);
   for (i=0; i<n; i++) {
      for (j=0; j<m; j++) {
         sum = AH(i,0)*Sinv(0,j);
         for (k=1; k<m; k++) sum += AH(i,k)*Sinv(k,j);
         S(i,j) = sum;
      }
   }
   for (i=0; i<n; i++) {
      for (j=0; j<i; j++) {
         sum = S(i,0)*AH(j,0);
         for (k=1; k<m; k++) sum += S(i,k)*AH(j,k);
         A(i,j) -= sum;
      }
      sum = S(i,0)*AH(i,0);
      for (k=1; k<m; k++) sum += S(i,k)*AH(i,k);
      A(i,i) -= sum;
   }*/

	  // return(LogDet(A,n));
}
   

double ComputeOmegaCI(MMatrix &A1, MMatrix &A2, int n,
                      MMatrix &B1, MMatrix &B2, int m,
                      MMatrix &H) 
{
   int i;
   double a=0.0, b=1.0, x, v, w, fx, ft, fv, fw, r, p, q, t,
        new_step, range, midpoint, tol;
   
   r = 0.5*(3-sqrt(5.0)); 

   v = a + r*(b-a);
   fv = LogDetOfUpdate(v,A1,A2,n,B1,B2,m,H);
   x = v;  w = v;
   fx=fv;  fw=fv;

   for (i=1;;i++) {  
      midpoint = 0.5*(a+b); range = b-a;
      tol = SQRT_EPSILON*fabs(x) + EPSILON/3;
      if ((fabs(x-midpoint) + 0.5*range <= 2*tol) || i>25) {
	 //  printf("Number of optimization steps: %d\n",i);
         if (x <= tol) return(0.0);
         if (1.0-x <= tol) return(1.0);
         return(x);
      }
      new_step = r*(x < midpoint ? b-x : a-x);
      if (fabs(x-w) >= tol) {
         t = (x-w) * (fx-fv);
         q = (x-v) * (fx-fw);
         p = (x-v)*q - (x-w)*t;
         q = 2*(q-t);
         if (q > 0) p = -p;
         else q = -q;
         if ((fabs(p) < fabs(new_step*q)) &&
             (p > q*(a-x+2*tol)) &&
             (p < q*(b-x-2*tol))) new_step = p/q;
      } 
      if (fabs(new_step) < tol) {
         if (new_step > 0) new_step = tol;
         else new_step = -tol;
      }
      t = x + new_step;
      ft = LogDetOfUpdate(t,A1,A2,n,B1,B2,m,H);
      if (ft <= fx) {
         if (t < x) b = x;
         else a = x;
         v = w;  w = x;  x = t;
         fv=fw;  fw=fx;  fx=ft;
      }
      else {
         if (t < x) a = t;
         else b = t;
         if (ft <= fw || w==x) {
            v = w;  w = t;
            fv=fw;  fw=ft;
         }
         else if( ft<=fv || v==x || v==w ) {
            v = t;
            fv = ft;
         }
      }
   }
}

double LogDetOfSum(double w,
                   MMatrix &A1, MMatrix &A2, 
                   MMatrix &B1, MMatrix &B2, int n)
{
   int i, j;
   MMatrix C(n,n); 
   double wi1, wi2;

   if ((w<EPSILON) || ((1-w)<EPSILON)) return(AlmostInfinite);
   wi1 = 1/w;  wi2 = 1/(1-w);

   C(0,0) = wi1*A1(0,0) + A2(0,0) + wi2*B1(0,0) + B2(0,0);
   for (i=1; i<n; i++) {
      for (j=0; j<i; j++) {
         C(i,j) = C(j,i) = wi1*A1(i,j) + A2(i,j) + 
                           wi2*B1(i,j) + B2(i,j); 
      }
      C(i,i) = wi1*A1(i,i) + A2(i,i) + 
               wi2*B1(i,i) + B2(i,i);
   }
   return(LogDet(C,n));
}

double ComputeOmegaCA(MMatrix &A1, MMatrix &A2, 
                      MMatrix &B1, MMatrix &B2, int n)
{
   int i;
   double a=0.0, b=1.0, x, v, w, fx, ft, fv, fw, r, p, q, t,
        new_step, range, midpoint, tol;
  
   r = 0.5*(3-sqrt(5.0));

   v = a + r*(b-a);
   fv = LogDetOfSum(v,A1,A2,B1,B2,n);
   x = v;  w = v;
   fx=fv;  fw=fv;

   for (i=1;;i++) {
      midpoint = 0.5*(a+b); range = b-a;
      tol = SQRT_EPSILON*fabs(x) + EPSILON/3;
      if ((fabs(x-midpoint) + 0.5*range <= 2*tol) || i>NUMIT) {
         printf("Number of optimization steps: %d\n",i);
         if (x <= tol) return(0.0);
         if (1.0-x <= tol) return(1.0);
         return(x);
      }
      new_step = r*(x < midpoint ? b-x : a-x);
      if (fabs(x-w) >= tol) {
         t = (x-w) * (fx-fv);
         q = (x-v) * (fx-fw);
         p = (x-v)*q - (x-w)*t;
         q = 2*(q-t);
         if (q > 0) p = -p;
         else q = -q;
         if ((fabs(p) < fabs(new_step*q)) &&
             (p > q*(a-x+2*tol)) &&
             (p < q*(b-x-2*tol))) new_step = p/q;
      }
      if (fabs(new_step) < tol) {
         if (new_step > 0) new_step = tol;
         else new_step = -tol;
      }
      t = x + new_step;
      ft = LogDetOfSum(t,A1,A2,B1,B2,n);
      if (ft <= fx) {
         if (t < x) b = x;
         else a = x;
         v = w;  w = x;  x = t;
         fv=fw;  fw=fx;  fx=ft;
      }
      else {
         if (t < x) a = t;
         else b = t;
         if (ft <= fw || w==x) {
            v = w;  w = t;
            fv=fw;  fw=ft;
         }
         else if( ft<=fv || v==x || v==w ) {
            v = t;
            fv = ft;
         }
      }
   }
}
      
