// Code for the grey system predictor

#include "grey.H"
#include "mmatrix.H"
#include "vvector.H"
#include "matrix.h"

#define TFORM 10.0  // used to transform the motion space into a more appropriate
                    // raw space otherwise, windows that change sign will
                    // have poor prediction results

#define SPEC_DENSITY 1

// setup storage for PVSM KF/EKF
GreyPred::GreyPred() : KalmanBase(), ExtendedKalmanBase(), _numwin(4), 
                      _posewin0(20), _flag(0), _plo(0), _phi(1), _ctr(0)
{
  // create for KF
  _RR = MMatrix(1,1);
  _HH = MMatrix(1,2);
  _QQ = MMatrix(2,2);
  _fundMat = MMatrix(2,2);

  for (int i = 0; i < 3; ++i) {
    _PP[i] = MMatrix(2,2); 
    _state[i] = MMatrix(2,1);
  }
  
  // create for EKF
  _ER = MMatrix(4,4);
  _EH = MMatrix(4,7);
  _EQ = MMatrix(7,7);
  _EfundMat = MMatrix(7,7);
  _EP = MMatrix(7,7);
  _Estate = MMatrix(7,1);
  _edt = _spd = 0.0;
}

// destructor
GreyPred::~GreyPred()
{
  _RR.deleteMatrix();
  _HH.deleteMatrix();
  _QQ.deleteMatrix();
  _fundMat.deleteMatrix();
  for (int i = 0; i < 3; ++i) {
    _PP[i].deleteMatrix(); 
    _state[i].deleteMatrix();
  }

  _ER.deleteMatrix();
  _EH.deleteMatrix();
  _EQ.deleteMatrix();
  _EfundMat.deleteMatrix();
  _EP.deleteMatrix();
  _Estate.deleteMatrix();

  for (int i = 0; i < _posewin0.num(); ++i)
     delete _posewin0[i];

}


void
GreyPred::prepare()
{
  // collect what we need from the bundle
  _numwin = _paramb->getGreyPts();
  _flag = 1;
  _plo = 0;
  _phi = 1;
  _ctr = 0;

   // if the Kalman flag is set we will Kalman filter
  if (_paramb->getKFGrey()) {
     double dt = 1.0 / _paramb->getCurSRate();
     double np = _paramb->getPosVar();
     double nq = _paramb->getQuatVar();
     double spd = SPEC_DENSITY;
  
     // initialize for KF
     _HH(0,0) = 1.0;
     _RR(0,0) = np;

     _fundMat(0,0) = _fundMat(1,1) = 1.0;
     _fundMat(0,1) = dt; 


     _QQ(0,0) = ((dt*dt*dt) / 3.0) * spd;
     _QQ(0,1) = _QQ(1,0) = ((dt*dt) / 2.0) * spd;
     _QQ(1,1) = dt * spd;

     for (int i = 0; i < 3; ++i) {
	_PP[i].setZeros();
	_PP[i](0,0) = _PP[i](1,1) = INITVAR;
	// zero out the state vectors
	_state[i](0,0) = _state[i](1,0) = 0.0;
     }
     
     // initialize EKF
     _edt = dt;
     _spd = spd;
     // we can set the measurment noise covariance matrix here (at least that can
     // stay constant	
     _ER.setZeros();
   
     for (int i = 0; i < 4; ++i)
      _ER(i,i) = nq;
     _EP.setZeros();
     for (int i = 0; i < 7; ++i) {
	if (i < 4)
	   _EP(i,i) = 1.0;
	else
	   _EP(i,i) = INITVAR;
	_Estate(i,0) = 0.0; 
     }  

  }

  // reset the window list
  for (int i = 0; i < _posewin0.num(); ++i)
     delete _posewin0[i]; 
  _posewin0.clear();
}

Pose *
GreyPred::predict(Pose *userpos, double pinterval)
{
  Pose *npos = 0;

  if (_paramb->getKFGrey()) {
     if (_ctr <= MAXBEGIN-6) 
       ++_ctr;
     if (_ctr == 1) {
	for (int i = 0; i < 3; ++i) 
	   (_state[i])(0,0) = userpos->_pos[i];
	for (int i = 0; i < 3; ++i)
	   _Estate(i,0) = userpos->_orient.V()[i];
	_Estate(3,0) = userpos->_orient.S();
        return userpos;
     }
      // Kalman Filter
      kalmanFilter(userpos, 1, 2, kPVSM);
      extendedKalmanFilter(userpos, 4, 7, kPVSM);
      if (_ctr < MAXBEGIN-6)
        return userpos; 
      else { // now start filling the pose records for prediction  
         Vec3d nv((_state[0])(0,0), (_state[1])(0,0), (_state[2])(0,0));
         Quat nq(_Estate(0,0), _Estate(1,0), _Estate(2,0), _Estate(3,0)); 
	 npos = new Pose(userpos->_tstamp, nv, normalize(nq));
      }
  }
 Pose *tmppos;
 if (_paramb->getKFGrey())
   tmppos = new Pose(npos);
 else
   tmppos = new Pose(userpos);
  
 // create and update the pose window
  if (_posewin0.num() < _numwin) {
    _posewin0 += tmppos;
    return userpos;
  }
  else {
    Pose *p;
    for (int i = 1; i < _numwin; ++i) {
       if (i == 1)
	  p =_posewin0[0];
      _posewin0[i-1] = _posewin0[i];
    }
    delete p;
    _posewin0.pop();
    _posewin0 += tmppos;

    return greyInterp(pinterval);  // predict
  }
}


// grey system predictor code, uses the interpolation scheme

Pose *
GreyPred::greyInterp(double pinterval)
{
  // compute the correct leadtime, with integer-based predictors
  // the lead time changes according to the current sampling rate
  if (_flag) {
    double csrate = _paramb->getCurSRate();
    double delta = 1.0 / csrate;  
    while (delta*(_phi) < pinterval) {
      _plo = _phi; 
      ++_phi;
    }
    _flag = 0;
  }

  Vec3d ppos(0.0,0.0,0.0);
  Quat  prot(0,0,0,1);
  double loq[4] = {0.0,0.0,0.0,0.0};
  double hiq[4] = {0.0,0.0,0.0,0.0};
  double csrate = _paramb->getCurSRate();
  double delta = 1.0 / csrate;
  FIGarray <Pose*> wintmp(_posewin0.num());
   
  for (int i = 0; i < _posewin0.num(); ++i) {
    Pose *tmp = new Pose(_posewin0[i]);
    tmp->_pos += Vec3d(TFORM,TFORM,TFORM);
    tmp->_orient.V() += Vec3d(TFORM,TFORM,TFORM);
    tmp->_orient.S() += TFORM;
    wintmp += tmp;  
  }
 
  FIGarray<Pose*> posewin1(_posewin0.num());
  
  for (int k = 1; k <= _posewin0.num(); ++k) {
    Pose *tmp = new Pose();
    for (int m = 0; m < k; ++m) {
      tmp->_pos = tmp->_pos + wintmp[m]->_pos;  
      tmp->_orient = tmp->_orient + wintmp[m]->_orient;
    }
    posewin1 += tmp;
  }

  // now we create the z1 variables from the differential equation
  FIGarray<Pose*> z1(posewin1.num()-1);

  for (int k = 1; k < posewin1.num(); ++k) {
    Pose *tmp = new Pose();
    tmp->_pos = 0.5 * (posewin1[k]->_pos + posewin1[k-1]->_pos);
    tmp->_orient = 0.5 * (posewin1[k]->_orient + posewin1[k-1]->_orient);
    z1 += tmp;
  }

  // estimate the a and b parameters for each pose component using least squares
  // approximation 

 
  for (int k = 0; k < 7; ++k) {
    // create the B matrix
    MMatrix *B = new MMatrix(z1.num(), 2);
    for (int i = 0; i < z1.num(); ++i) {
      if (k < 3) 
       (*B)(i,0) = -z1[i]->_pos[k];
      else if ((k > 2) && (k < 6)) 
       (*B)(i,0) = -z1[i]->_orient.V()[k-3];
      else 
       (*B)(i,0) = -z1[i]->_orient.S();
     (*B)(i,1) = 1.0;
    }
    

    MMatrix *BT = B->transpose();
    MMatrix *Btmp = mmult(BT, B); //square matrix
    // invert Btmp

    int m,n; Btmp->getDim(&m,&n);
    
    MMatrix *IB = new MMatrix(n,n);
    matrix M1(n,n);
    for (int r = 0; r < n ; ++r)
      for (int s = 0; s < n; ++s)
	M1.setvalue(r,s, (*Btmp)(r,s)); 
    
    M1.invert();
    for (int r = 0; r < n; ++r)
      for (int s = 0; s < n; ++s) {
     double g;
     int p;
     M1.getvalue(r,s, g, p);  
     (*IB)(r,s) = g;
    }
    
    MMatrix *Btmp2 = mmult(IB, BT);

    Btmp2->getDim(&m,&n);
    MMatrix *colvec = new MMatrix(n,1);
    for (int r = 1; r < wintmp.num(); ++r) {
      if (k < 3)
	(*colvec)(r-1,0) = wintmp[r]->_pos[k];
      else if ((k > 2) && (k < 6))        
        (*colvec)(r-1,0) = wintmp[r]->_orient.V()[k-3];
      else
        (*colvec)(r-1,0) = wintmp[r]->_orient.S();
    }
 
    // get params a and b
    MMatrix *params = mmult(Btmp2, colvec);  // should give us a 2 by 1 matrix
      double initpt;
    // find the intial point for the IAGO
     if (k < 3)
	initpt = wintmp[0]->_pos[k];
     else if ((k > 2) && (k < 6))
        initpt = wintmp[0]->_orient.V()[k-3];
      else
        initpt = wintmp[0]->_orient.S();

    double lopred, hipred, actpred;
    if (pinterval  == (_phi * delta)) {
    // now compute the predicted value 
      double parta = ((initpt - ((*params)(1,0) / (*params)(0,0))) *
                   exp(-(*params)(0,0) * ((_phi-1)+_numwin))) + ((*params)(1,0) / 
                   (*params)(0,0));
    
      double partb = ((initpt - ((*params)(1,0) / (*params)(0,0))) *
                exp(-(*params)(0,0) * ((_phi-1)+_numwin-1))) + ((*params)(1,0) / 
                (*params)(0,0));
      actpred = (parta - partb)- TFORM;

       if ((k > 2) && (k < 6)) 
        prot.V()[k-3] = actpred;
      else if (k == 6) {
        prot.S() = actpred;
        normalize(prot);
      } 
     
    }
    else {  // get the low and high predictions
      if ((_plo - 1) == -1) {
	if (k < 3)
	  lopred = _posewin0[_numwin-1]->_pos[k];
	else if ((k > 2) && (k < 6))
	  lopred = _posewin0[_numwin-1]->_orient.V()[k-3];
	else
	  lopred = _posewin0[_numwin-1]->_orient.S();
      }       
      else {
	double parta = ((initpt - ((*params)(1,0) / (*params)(0,0))) *
			exp(-(*params)(0,0) * ((_plo-1)+_numwin))) + 
	               ((*params)(1,0) / (*params)(0,0));
    
	double partb = ((initpt - ((*params)(1,0) / (*params)(0,0))) *
                        exp(-(*params)(0,0) * ((_plo-1)+_numwin-1))) + 
                        ((*params)(1,0) / (*params)(0,0));
	lopred = (parta - partb) - TFORM;
      }
       double parta = ((initpt - ((*params)(1,0) / (*params)(0,0))) *
		     exp(-(*params)(0,0) * ((_phi-1)+_numwin))) + 
                     ((*params)(1,0) / (*params)(0,0)); 
    
       double partb = ((initpt - ((*params)(1,0) / (*params)(0,0))) *
                       exp(-(*params)(0,0) * ((_phi-1)+_numwin-1))) + 
                       ((*params)(1,0) / (*params)(0,0));
       hipred = (parta - partb) - TFORM;
 

       // interpolation code here
       if (k < 3) {
	 double t2 = (_phi * (1.0 / csrate));
	 double t1 = (_plo * (1.0 / csrate));
	 actpred = (((hipred - lopred) / (t2 - t1)) * (pinterval - t1)) + lopred;
       }
       else {
	 loq[k-3] = lopred;  // need to store quaterion values so we can 
	 hiq[k-3] = hipred;  // use quaternion interpolation
       } 
    }
    
    if (k < 3)
      ppos[k] = actpred;
   
    // delete matricies
    // M1.destroy();
    delete B;
    delete BT;
    delete IB;
    delete Btmp;
    delete Btmp2;
    delete colvec;
    delete params;
  }

  double t2 = (_phi * (1.0 / csrate));
   double t1 = (_plo * (1.0 / csrate));
   if (pinterval  != (_phi * delta)) { // loq and hiq have been filled
     prot = slerp(normalize(Quat(loq[0],loq[1],loq[2],loq[3])), 
		  normalize(Quat(hiq[0],hiq[1],hiq[2],hiq[3])), 
		  (pinterval-t1) / (t2 - t1));
   }
  
  // delete extra storage
  for (int i = 0; i < wintmp.num(); ++i) 
    delete wintmp[i];
  for (int i = 0; i < posewin1.num(); ++i)
    delete posewin1[i];  
  for (int i = 0; i < z1.num(); ++i)
    delete z1[i];

  return new Pose(_posewin0[_numwin-1]->_tstamp + pinterval, ppos, normalize(prot));
}






// old grey prediction code
// finds the integerized prediction time closest to the actual prediction time


Pose *
GreyPred::greyNeighbor(double pinterval)
{

  // compute the correct leadtime, with integer-based predictors
  // the lead time changes according to the current sampling rate
   double csrate = _paramb->getCurSRate();
   double origsr = _paramb->getOrigSRate();
   int leadtime = _paramb->getLeadTime();
   double factor =  origsr / csrate;
   double tmpld = (double)leadtime / factor;
   if ((tmpld - floor(tmpld) < 0.5))
    leadtime = (int)floor(tmpld);
   else
    leadtime = (int)ceil(tmpld); 

   if (_flag) {
     cerr << "Lead: "<< _paramb->getLeadTime()/(origsr / csrate) << endl;
     cerr << "Lead int " << leadtime << endl;
     double tmp;
     if (leadtime == 0)
       tmp = 1;
     else
       tmp = leadtime;
     cerr << "Desired P Time: " << 1000.0 * pinterval << endl;
     cerr << "Actual P Time: " << 1000.0 * (tmp) * (1.0 / csrate) << endl << endl;
     _flag = 0;
   }  
   

   Vec3d ppos(0.0,0.0,0.0);
   Quat  prot(0,0,0,1);
   FIGarray <Pose*> wintmp(_posewin0.num());
   
   for (int i = 0; i < _posewin0.num(); ++i) {
     Pose *tmp = new Pose(_posewin0[i]);
     tmp->_pos += Vec3d(TFORM,TFORM,TFORM);
     tmp->_orient.V() += Vec3d(TFORM,TFORM,TFORM);
     tmp->_orient.S() += TFORM;
     wintmp += tmp;  
   }
 
   FIGarray<Pose*> posewin1(_posewin0.num());
  
   for (int k = 1; k <= _posewin0.num(); ++k) {
     Pose *tmp = new Pose();
     for (int m = 0; m < k; ++m) {
       tmp->_pos = tmp->_pos + wintmp[m]->_pos;  
       tmp->_orient = tmp->_orient + wintmp[m]->_orient;
     }
     posewin1 += tmp;
   }

  // now we create the z1 variables from the differential equation
   FIGarray<Pose*> z1(posewin1.num()-1);

   for (int k = 1; k < posewin1.num(); ++k) {
     Pose *tmp = new Pose();
     tmp->_pos = 0.5 * (posewin1[k]->_pos + posewin1[k-1]->_pos);
     tmp->_orient = 0.5 * (posewin1[k]->_orient + posewin1[k-1]->_orient);
     z1 += tmp;
   }

  // estimate the a and b parameters for each pose component using least squares
  // approximation 

  
   for (int k = 0; k < 7; ++k) {
    // create the B matrix
     MMatrix *B = new MMatrix(z1.num(), 2);
     for (int i = 0; i < z1.num(); ++i) {
       if (k < 3) 
	 (*B)(i,0) = -z1[i]->_pos[k];
       else if ((k > 2) && (k < 6)) 
	 (*B)(i,0) = -z1[i]->_orient.V()[k-3];
       else 
	 (*B)(i,0) = -z1[i]->_orient.S();
       (*B)(i,1) = 1.0;
     }
         

     MMatrix *BT = B->transpose();
     MMatrix *Btmp = mmult(BT, B); //square matrix
    // invert Btmp
 
     int m,n; Btmp->getDim(&m,&n);
    
     MMatrix *IB = new MMatrix(n,n);
     matrix M1(n,n);
     for (int r = 0; r < n ; ++r)
      for (int s = 0; s < n; ++s)
   	M1.setvalue(r,s, (*Btmp)(r,s)); 
     
     M1.invert();
     for (int r = 0; r < n; ++r)
      for (int s = 0; s < n; ++s) {
     double g;
     int p;
     M1.getvalue(r,s, g, p);  
     (*IB)(r,s) = g;
    }
    
    MMatrix *Btmp2 = mmult(IB, BT);

    Btmp2->getDim(&m,&n);
    MMatrix *colvec = new MMatrix(n,1);
    for (int r = 1; r < wintmp.num(); ++r) {
      if (k < 3)
	(*colvec)(r-1,0) = wintmp[r]->_pos[k];
      else if ((k > 2) && (k < 6))        
        (*colvec)(r-1,0) = wintmp[r]->_orient.V()[k-3];
      else
        (*colvec)(r-1,0) = wintmp[r]->_orient.S();
    }
     
    // get params a and b
    MMatrix *params = mmult(Btmp2, colvec);  // should give us a 2 by 1 matrix
    double initpt;
    // find the intial point for the IAGO
    if (k < 3)
      initpt = wintmp[0]->_pos[k];
    else if ((k > 2) && (k < 6))
        initpt = wintmp[0]->_orient.V()[k-3];
    else
      initpt = wintmp[0]->_orient.S();
    // now compute the predicted value 

    double parta = ((initpt - ((*params)(1,0) / (*params)(0,0))) *
                   exp(-(*params)(0,0) * (leadtime+_numwin))) + ((*params)(1,0) / (*params)(0,0));
    
    double partb = ((initpt - ((*params)(1,0) / (*params)(0,0))) *
               exp(-(*params)(0,0) * (leadtime+_numwin-1))) + ((*params)(1,0) / (*params)(0,0));
        
    double actpred = (parta - partb)- TFORM;
    if (k < 3)
      ppos[k] = actpred;
    else if ((k > 2) && (k < 6))
      prot.V()[k-3] = actpred;
    else
      prot.S() = actpred;

    // delete matricies
    M1.destroy();
    delete B;
    delete BT;
    delete IB;
    delete Btmp;
    delete Btmp2;
    delete colvec;
    delete params;
  }
    
  // delete extra storage
  for (int i = 0; i < wintmp.num(); ++i) 
    delete wintmp[i];
  for (int i = 0; i < posewin1.num(); ++i)
    delete posewin1[i];  
  for (int i = 0; i < z1.num(); ++i)
    delete z1[i];

  return new Pose(_posewin0[_numwin-1]->_tstamp + pinterval, ppos, normalize(prot));
}
