// MMAE 3 model predictor

#include "mmae3model.H"

#define EP 0.01
#define M 3 // length of measurement vector
#define QM 4 // length of quaternion
#define LESSNUM 6

MMAE3Model::MMAE3Model() : _ctr(0)
{
   // create the three models for KF
   _kfmod[0] = MMAEKalmanBase(kPSM);
   _kfmod[1] = MMAEKalmanBase(kPVSM);
   _kfmod[2] = MMAEKalmanBase(kPVASM);

   _kfprob[0] = _kfprob[1] = _kfprob[2] = 0.0;
   _kfprobtmp[0] = _kfprobtmp[1] = _kfprobtmp[2] = 0.0;

   // create the three models for EKF
   _ekfmod[0] = MMAEExtendedKalmanBase(kPSM);
   _ekfmod[1] = MMAEExtendedKalmanBase(kPVSM);
   _ekfmod[2] = MMAEExtendedKalmanBase(kPVASM);

   _ekfprob[0] = _ekfprob[1] = _ekfprob[2] = 0.0;
   _ekfprobtmp[0] = _ekfprobtmp[1] = _ekfprobtmp[2] = 0.0;
}


void
MMAE3Model::initialize(double specd, double dt, double pn, 
                       double qn, CombinerType cmb)
{
 _ctr = 0;
 _cmb = cmb;

 // initialize each KF model
 _kfmod[0].initialize(specd, dt, pn);
 _kfmod[1].initialize(specd, dt, pn);
 _kfmod[2].initialize(specd, dt, pn);  

 // initialize the conditional probabilities for the kf models 
 _kfprob[0] = _kfprob[1] = _kfprob[2] = (1.0 / 3.0); 

 // initialize each EKF model
 _ekfmod[0].initialize(specd, dt, qn);
 _ekfmod[1].initialize(specd, dt, qn);
 _ekfmod[2].initialize(specd, dt, qn);  

 _ekfprob[0] = _ekfprob[1] = _ekfprob[2] = (1.0 / 3.0); 
  
}


int MMAE3Model::getHighestModelKF()
{
  int highest = 0;
   
  for (int i = 1; i < 3; ++i)
   if (_kfprob[i] > _kfprob[highest])
     highest = i;
 
 return highest;
}

int MMAE3Model::getHighestModelEKF()
{
  int highest = 0;
   
  for (int i = 1; i < 3; ++i)
   if (_ekfprob[i] > _ekfprob[highest])
     highest = i;
 
 return highest;
}


void
MMAE3Model::computeCondProbKF()
{
 for (int k = 0; k < 3; ++k) {
    // compute the kth f function
    double a = (transp(_kfmod[k].getResidual()) * 
                (minverse2(_kfmod[k].getResidualCov()) * 
                _kfmod[k].getResidual()))(0,0);
    double fk = 1.0 / (pow(2.0*M_PI, M/2.0) * sqrt(det(_kfmod[k].getResidualCov()))) *
                 exp(-0.5 * a);
    double num = fk * _kfprob[k];
    
    // compute the denominator sum
    double denum = 0.0;
    for (int j = 0; j < 3; ++j) {
     double b = (transp(_kfmod[j].getResidual()) * 
                (minverse2(_kfmod[j].getResidualCov()) * 
                 _kfmod[j].getResidual()))(0,0);
     denum += (1.0 / (pow(2.0*M_PI, M/2.0) * sqrt(det(_kfmod[j].getResidualCov()))) *
	       exp(-0.5 * b)) * _kfprob[j];
    }
    // compute new conditional probability 
    _kfprobtmp[k] = num / denum;
 }
// cerr << "KF: ";
 // make lower bounds on probs 
 for (int k = 0; k < 3; ++k) {
//    cerr << _kfprobtmp[k] << " ";
    _kfprob[k] = _kfprobtmp[k];  
    if (_kfprob[k] < EP)
       _kfprob[k] = EP;
 }
// cerr << endl;
}


void
MMAE3Model::computeCondProbEKF()
{
 for (int k = 0; k < 3; ++k) {
    // compute the kth f function
    //    transp(_ekfmod[k].getResidual()).printMatrix();
    //  (_ekfmod[k].getResidualCov()).printMatrix();
    //  minverse2(_ekfmod[k].getResidualCov()).printMatrix();
 
    double a = (transp(_ekfmod[k].getResidual()) * 
                (minverse2(_ekfmod[k].getResidualCov()) * 
                _ekfmod[k].getResidual()))(0,0);
    double fk = 1.0 / (pow(2.0*M_PI, QM/2.0) * sqrt(det(_ekfmod[k].getResidualCov()))) *
                 exp(-0.5 * a);
    double num = fk * _ekfprob[k];
    
    // compute the denominator sum
    double denum = 0.0;
    for (int j = 0; j < 3; ++j) {
     double b = (transp(_ekfmod[j].getResidual()) * 
                (minverse2(_ekfmod[j].getResidualCov()) * 
                 _ekfmod[j].getResidual()))(0,0);
     denum += (1.0 / (pow(2.0*M_PI, QM/2.0) * sqrt(det(_ekfmod[j].getResidualCov()))) *
	       exp(-0.5 * b)) * _ekfprob[j];
    }
    // compute new conditional probability 
    _ekfprobtmp[k] = num / denum;
 }
// cerr << "EKF: ";
 // make lower bounds on probs 
 for (int k = 0; k < 3; ++k) {
    //   cerr << _ekfprobtmp[k] << " ";
    _ekfprob[k] = _ekfprobtmp[k];  
    if (_ekfprob[k] < EP)
       _ekfprob[k] = EP;
 }
// cerr << endl;
}



Vec3d
MMAE3Model::predictPos(double pinterval)
{
  Vec3d ppos(0.0,0.0,0.0);

  if (_cmb == tMAX) {
    // find the highest probability and use the coresponding state vector
    // and kf model to predict
    int modnum = getHighestModelKF();    
    if (modnum == 0) { // PSM model has highest probability
       // make predicton	
      for (int i = 0; i < 3; ++i)
        ppos[i] = _kfmod[0].getStateVector()(i,0);
      // restart other filters
      if (_kfprobtmp[1] < EP) {
	 for (int j = 0; j < 3; ++j) {
           _kfmod[1].getStateVector()(j*2,0) = ppos[j];
	   //   _kfmod[1].getCov()(j*2,j*2) = _kfmod[0].getCov()(j,j);
	 }
      }  
      if (_kfprobtmp[2] < EP) {
         for (int j = 0; j < 3; ++j) {
           _kfmod[2].getStateVector()(j*3,0) = ppos[j];
	   // _kfmod[2].getCov()(j*3,j*3) = _kfmod[0].getCov()(j,j);
         }
      }
    }
    else if (modnum == 1) { // PVSM model has highest probability
       // make prediction
       MMatrix pmat = _kfmod[1].getFundMat();
       pmat(0,1) = pmat(2,3) = pmat(4,5) = pinterval;
       MMatrix tmp = pmat * _kfmod[1].getStateVector();
       for(int i = 0; i < 3; ++i)
	  ppos[i] = tmp(i*2,0); 
       // restart other filters
       if (_kfprobtmp[0] < EP) {
	  for (int j = 0; j < 3; ++j) {
            _kfmod[0].getStateVector()(j,0) = _kfmod[1].getStateVector()(j*2,0);
	    // _kfmod[0].getCov()(j,j) = _kfmod[1].getCov()(j*2,j*2);
          }
       }
       if (_kfprobtmp[2] < EP) {
	 for (int j = 0; j < 3; ++j) {
	   _kfmod[2].getStateVector()(j*3,0) = _kfmod[1].getStateVector()(j*2,0);
           _kfmod[2].getStateVector()((j*3)+1,0) = _kfmod[1].getStateVector()((j*2)+1,0);
	   // _kfmod[2].getCov()(j*3,j*3) = _kfmod[1].getCov()(j*2,j*2);
	   // _kfmod[2].getCov()((j*3)+1,(j*3)+1) = _kfmod[1].getCov()((j*2)+1,(j*2)+1);
          }
       }
    }
    else {  // PVASM has highest probability
       // make prediction
      MMatrix pmat = _kfmod[2].getFundMat(); 
      pmat(0,1) = pmat(1,2) = pmat(3,4) = 
      pmat(4,5) = pmat(6,7) = pmat(7,8) = pinterval;
      pmat(0,2) = pmat(3,5) = pmat(6,8) = pinterval*pinterval/2.0;
      MMatrix tmp = pmat * _kfmod[2].getStateVector();
      for(int i = 0; i < 3; ++i)
	  ppos[i] = tmp(i*3,0); 
      // restart other filters
      if (_kfprobtmp[0] < EP) {
         for (int j = 0; j < 3; ++j) {
	   _kfmod[0].getStateVector()(j,0) = _kfmod[2].getStateVector()(j*3,0);
	   // _kfmod[0].getCov()(j,j) = _kfmod[2].getCov()(j*3,j*3);
         }
      }
      if (_kfprobtmp[1] < EP) {
	for (int j = 0; j < 3; ++j) {
	  _kfmod[1].getStateVector()(j*2,0) = _kfmod[2].getStateVector()(j*3,0); 
          _kfmod[1].getStateVector()((j*2)+1,0) = _kfmod[2].getStateVector()((j*3)+1,0); 
	  // _kfmod[1].getCov()(j*2,j*2) = _kfmod[2].getCov()(j*3,j*3);
	  // _kfmod[1].getCov()((j*2)+1,(j*2)+1) = _kfmod[2].getCov()((j*3)+1,(j*3)+1);
	}
      }
    }
 }
 else if (_cmb == tSUM) {
    // there a number of ways to do this
    // add up the position

    Vec3d pp1(0.0,0.0,0.0);
    Vec3d pp2(0.0,0.0,0.0);
    Vec3d pp3(0.0,0.0,0.0);
    
    // add up velocities
    Vec3d v1(0.0,0.0,0.0);
    Vec3d v2(0.0,0.0,0.0);

    for (int i = 0; i < 3; ++i) {
      pp1[i] = _kfmod[0].getStateVector()(i,0);
      pp2[i] = _kfmod[1].getStateVector()(i*2,0);
      pp3[i] = _kfmod[2].getStateVector()(i*3,0);
      v1[i]  = _kfmod[1].getStateVector()((i*2)+1,0);
      v2[i]  = _kfmod[2].getStateVector()((i*3)+1,0); 
    }
     
    // blend
    Vec3d pp4 = _kfprobtmp[0] * pp1 + _kfprobtmp[1] * pp2 + _kfprobtmp[2] * pp3;
    Vec3d v3 = (_kfprobtmp[1] / (_kfprobtmp[1] + _kfprobtmp[2])) * v1 + 
               (_kfprobtmp[2] / (_kfprobtmp[1] + _kfprobtmp[2])) * v2;

    for (int i = 0; i < 3; ++i) {
      if (_kfprobtmp[0] < EP)
	 _kfmod[0].getStateVector()(i,0) = pp4[i];
      if (_kfprobtmp[1] < EP) {
	 _kfmod[1].getStateVector()(i*2,0) = pp4[i];
         _kfmod[1].getStateVector()((i*2)+1,0) = v3[i];
      }
      if (_kfprobtmp[2] < EP) {
	 _kfmod[2].getStateVector()(i*3,0) = pp4[i];
	 _kfmod[2].getStateVector()((i*3)+1,0) = v3[i];
      }
    }
    
    MMatrix nstate(6,1);
    for (int i = 0; i < 3; ++i) {
      nstate(i*2,0) = pp4[i];
      nstate((i*2)+1, 0) = v3[i];
    }

    MMatrix pmat = _kfmod[1].getFundMat();
    pmat(0,1) = pmat(2,3) = pmat(4,5) = pinterval;
    MMatrix tmp = pmat * nstate;
    for(int i = 0; i < 3; ++i)
       pp2[i] = tmp(i*2,0); 
    
    for (int i = 0; i < 3; ++i)
      ppos[i] = pp2[i]; 
 }

 return ppos;
}


Quat
MMAE3Model::predictQuat(double pinterval)
{
  Quat prot(0.0,0.0,0.0,1.0);

  if (_cmb == tMAX) {
    // find the highest probability and use the coresponding state vector
    // and ekf model to predict
    int modnum = getHighestModelEKF();    
    if (modnum == 0) { // PSM model has highest probability
       // make predicton	
      for (int i = 0; i < 3; ++i)
        prot.V()[i] = _ekfmod[0].getStateVector()(i,0);
      prot.S() =_ekfmod[0].getStateVector()(3,0); 

      // restart other filters
      if (_ekfprobtmp[1] < EP) {
	 for (int j = 0; j < 4; ++j) {
	   _ekfmod[1].getStateVector()(j,0) = _ekfmod[0].getStateVector()(j,0);
	   //  _ekfmod[1].getCov()(j,j) = _ekfmod[0].getCov()(j,j);
	 }
      }  
      if (_ekfprobtmp[2] < EP) {
         for (int j = 0; j < 4; ++j) {
           _ekfmod[2].getStateVector()(j,0) = _ekfmod[0].getStateVector()(j,0);
	   //  _ekfmod[2].getCov()(j,j) = _ekfmod[0].getCov()(j,j);
         }
      }
    }
    else if (modnum == 1) { // PVSM model has highest probability
       // make prediction
        MMatrix estate = _ekfmod[1].getStateVector();
        Quat tq(estate(0,0), estate(1,0), estate(2,0), estate(3,0));
        Quat tw(estate(4,0), estate(5,0), estate(6,0), 0);
        prot = tq + (pinterval * 0.5 * (tq * tw));
       // restart other filters
      if (_ekfprobtmp[0] < EP) {
	  for (int j = 0; j < 4; ++j) {
            _ekfmod[0].getStateVector()(j,0) = _ekfmod[1].getStateVector()(j,0);
	    //  _ekfmod[0].getCov()(j,j) = _ekfmod[1].getCov()(j,j);
          }
       }
       if (_ekfprobtmp[2] < EP) {
	 for (int j = 0; j < 7; ++j) {
	   _ekfmod[2].getStateVector()(j,0) = _ekfmod[1].getStateVector()(j,0);
	   // _ekfmod[2].getCov()(j,j) = _ekfmod[1].getCov()(j,j);
         }
       }
    }
    else {  // PVASM has highest probability
      // make prediction
      MMatrix estate = _ekfmod[2].getStateVector();
      Quat tq(estate(0,0), estate(1,0), estate(2,0), estate(3,0));
      Quat tw(estate(4,0), estate(5,0), estate(6,0), 0);
      Quat ta(estate(7,0), estate(8,0), estate(9,0), 0);
      Quat b = (pinterval * 0.5) * tq * tw;
      Quat c = ((-1.0 / 8.0) * pinterval * pinterval * 
	        tw.length()  *
	        tw.length()) * tq;
      Quat d = (0.25 * pinterval * pinterval) * tq * ta; 
      prot = tq + b + c + d;      

      // restart other filters
      if (_ekfprobtmp[0] < EP) {
         for (int j = 0; j < 4; ++j) {
	   _ekfmod[0].getStateVector()(j,0) = _ekfmod[2].getStateVector()(j,0);
	   // _ekfmod[0].getCov()(j,j) = _ekfmod[2].getCov()(j,j);
         }
      }
      if (_ekfprobtmp[1] < EP) {
	for (int j = 0; j < 7; ++j) {
	  _ekfmod[1].getStateVector()(j,0) = _ekfmod[2].getStateVector()(j,0); 
	  //  _ekfmod[1].getCov()(j,j) = _ekfmod[2].getCov()(j,j);
	}
      }
    }
  }
  else if (_cmb == tSUM) {
    // there a number of ways to do this
    // compute weighted sums of quats and ang vel's

    Quat q1(0.0,0.0,0.0,1.0);
    Quat q2(0.0,0.0,0.0,1.0);
    Quat q3(0.0,0.0,0.0,1.0);
    
    // add up velocities
    Quat av1(0.0,0.0,0.0,0.0);
    Quat av2(0.0,0.0,0.0,0.0);

    for (int i = 0; i < 3; ++i) {
      q1.V()[i] = _ekfmod[0].getStateVector()(i,0);
      q2.V()[i] = _ekfmod[1].getStateVector()(i,0);
      q3.V()[i] = _ekfmod[2].getStateVector()(i,0);
      av1.V()[i]  = _ekfmod[1].getStateVector()(i+4,0);
      av2.V()[i]  = _ekfmod[2].getStateVector()(i+4,0); 
    }
    q1.S() = _ekfmod[0].getStateVector()(3,0);
    q2.S() = _ekfmod[1].getStateVector()(3,0);
    q3.S() = _ekfmod[2].getStateVector()(3,0);
     
    // blend
    Quat q4 = _ekfprobtmp[0] * q1 + _ekfprobtmp[1] * q2 + _ekfprobtmp[2] * q3;
    Quat av3 = (_ekfprobtmp[1] / (_ekfprobtmp[1] + _ekfprobtmp[2])) * av1 + 
               (_ekfprobtmp[2] / (_ekfprobtmp[1] + _ekfprobtmp[2])) * av2;

    for (int i = 0; i < 3; ++i) {
       if (_ekfprobtmp[0] < EP) {
	 _ekfmod[0].getStateVector()(i,0) = q4.V()[i];
         if (i == 0)
           _ekfmod[0].getStateVector()(3,0) = q4.S();
       }
      if (_ekfprobtmp[1] < EP) {
	 _ekfmod[1].getStateVector()(i,0) = q4.V()[i];
         if (i == 0)
           _ekfmod[1].getStateVector()(3,0) = q4.S();
         _ekfmod[1].getStateVector()(i+4,0) = av3.V()[i];
      }
      if (_ekfprobtmp[2] < EP) {
	 _ekfmod[2].getStateVector()(i,0) = q4.V()[i];
         if (i == 0)
           _ekfmod[2].getStateVector()(3,0) = q4.S();
	 _ekfmod[2].getStateVector()(i+4,0) = av3.V()[i];
      }
    }
    
    MMatrix estate(7,1);
    for (int i = 0; i < 3; ++i) {
      estate(i,0) = q4.V()[i];
      estate(i+4, 0) = av3.V()[i];
    }
    estate(3,0) = q4.S();

    Quat tq(estate(0,0), estate(1,0), estate(2,0), estate(3,0));
    Quat tw(estate(4,0), estate(5,0), estate(6,0), 0);
    prot = tq + (pinterval * 0.5 * (tq * tw));
  }  


  return prot;
}



Pose * 
MMAE3Model::filter(Pose *userpos, double pinterval)
{
 if (_ctr <= MAXBEGIN-LESSNUM)
   ++_ctr;

 if (_ctr == 1) { // initialize state vectors 
    for (int i = 0; i < 3; ++i) 
       for (int j = 0; j < 3; ++j) {
	 _kfmod[i].getStateVector()(j*(i+1),0) = userpos->_pos[j];
         _ekfmod[i].getStateVector()(j,0) = userpos->_orient.V()[j];
       }
    for (int i = 0; i < 3; ++i)
      _ekfmod[i].getStateVector()(3,0) = userpos->_orient.S();
    return userpos;
 }  
 
 // run each filter, both KF and EKF
 for (int i = 0; i < 3; ++i) {
    _kfmod[i].mmaeKalmanFilter(userpos, 3);
    _ekfmod[i].mmaeExtendedKalmanFilter(userpos, 4);
 }

 // compute the conditional probabilities
 computeCondProbKF();
 computeCondProbEKF();
   
 if (_ctr < MAXBEGIN-LESSNUM) // let it run for a little bit before we predict
   return userpos;


 // predict and reinitialize if needed
 Vec3d ppos = predictPos(pinterval);
 Quat prot = predictQuat(pinterval);

 return new Pose(userpos->_tstamp + pinterval, ppos, normalize(prot)); 
}
