// MMAE single model (PVSM), multiparameter predictor

#include "mmaemultip.H"

#define EP 0.01
#define M 3 // length of measurement vector
#define QM 4 // length of quaternion
#define LESSNUM 6

MMAEMultiParam::MMAEMultiParam() : _ctr(0)
{
   // create the KF three models
   _kfmod[0] = MMAEKalmanBase(kPVSM);
   _kfmod[1] = MMAEKalmanBase(kPVSM);
   _kfmod[2] = MMAEKalmanBase(kPVSM);

   _kfprob[0] = _kfprob[1] = _kfprob[2] = 0.0;
   _kfprobtmp[0] = _kfprobtmp[1] = _kfprobtmp[2] = 0.0;

   _ekfmod[0] = MMAEExtendedKalmanBase(kPVSM);
   _ekfmod[1] = MMAEExtendedKalmanBase(kPVSM);
   _ekfmod[2] = MMAEExtendedKalmanBase(kPVSM);

   _ekfprob[0] = _ekfprob[1] = _ekfprob[2] = 0.0;
   _ekfprobtmp[0] = _ekfprobtmp[1] = _ekfprobtmp[2] = 0.0;
}


void
MMAEMultiParam::initialize(double specd, double sep, double dt, double pn, 
                          double qn, CombinerType cmb)
{
 _ctr = 0;
 _cmb = cmb;

 // initialize each KF model
 _kfmod[0].initialize(specd, dt, pn);
 _kfmod[1].initialize(specd+sep, dt, pn);
 _kfmod[2].initialize(specd+(2.0*sep), dt, pn);  

 // initialize the conditional probabilities for the kf models 
 _kfprob[0] = _kfprob[1] = _kfprob[2] = (1.0 / 3.0); 

 // initialize each EKF model
 _ekfmod[0].initialize(specd, dt, qn);
 _ekfmod[1].initialize(specd+sep, dt, qn);
 _ekfmod[2].initialize(specd+(2.0*sep), dt, qn);  

 // initialize the conditional probabilities for the ekf models 
 _ekfprob[0] = _ekfprob[1] = _ekfprob[2] = (1.0 / 3.0); 

  
}


int MMAEMultiParam::getHighestModelKF()
{
  int highest = 0;
   
  for (int i = 1; i < 3; ++i)
   if (_kfprob[i] > _kfprob[highest])
     highest = i;
 
 return highest;
}


int MMAEMultiParam::getHighestModelEKF()
{
  int highest = 0;
   
  for (int i = 1; i < 3; ++i)
   if (_ekfprob[i] > _ekfprob[highest])
     highest = i;
 
 return highest;
}


void MMAEMultiParam::computeCondProbKF()
{
  // compute the conditional probabilities
 for (int k = 0; k < 3; ++k) {
    // compute the kth f function
    double a = (transp(_kfmod[k].getResidual()) * 
                (minverse2(_kfmod[k].getResidualCov()) * 
                _kfmod[k].getResidual()))(0,0);
    double fk = 1.0 / (pow(2.0*M_PI, M/2.0) * sqrt(det(_kfmod[k].getResidualCov()))) *
                 exp(-0.5 * a);
    double num = fk * _kfprob[k];
    
    // compute the denominator sum
    double denum = 0.0;
    for (int j = 0; j < 3; ++j) {
     double b = (transp(_kfmod[j].getResidual()) * 
                (minverse2(_kfmod[j].getResidualCov()) * 
                 _kfmod[j].getResidual()))(0,0);
     denum += (1.0 / (pow(2.0*M_PI, M/2.0) * sqrt(det(_kfmod[j].getResidualCov()))) *
	       exp(-0.5 * b)) * _kfprob[j];
    }
    // compute new conditional probability 
    _kfprobtmp[k] = num / denum;
 }
 // make lower bounds on probs 
// cerr << "KF: ";
 for (int k = 0; k < 3; ++k) {
    //   cerr << _kfprobtmp[k] << " ";
    _kfprob[k] = _kfprobtmp[k];  
    if (_kfprob[k] < EP)
       _kfprob[k] = EP;
 }
// cerr << endl;

}

void MMAEMultiParam::computeCondProbEKF()
{
  // compute the conditional probabilities
 for (int k = 0; k < 3; ++k) {
    // compute the kth f function
    double a = (transp(_ekfmod[k].getResidual()) * 
                (minverse2(_ekfmod[k].getResidualCov()) * 
                _ekfmod[k].getResidual()))(0,0);
    double fk = 1.0 / (pow(2.0*M_PI, QM/2.0) * sqrt(det(_ekfmod[k].getResidualCov()))) *
                 exp(-0.5 * a);
    double num = fk * _ekfprob[k];
    
    // compute the denominator sum
    double denum = 0.0;
    for (int j = 0; j < 3; ++j) {
     double b = (transp(_ekfmod[j].getResidual()) * 
                (minverse2(_ekfmod[j].getResidualCov()) * 
                 _ekfmod[j].getResidual()))(0,0);
     denum += (1.0 / (pow(2.0*M_PI, QM/2.0) * sqrt(det(_ekfmod[j].getResidualCov()))) *
	       exp(-0.5 * b)) * _ekfprob[j];
    }
    // compute new conditional probability 
    _ekfprobtmp[k] = num / denum;
 }
 // make lower bounds on probs 
// cerr << "EKF: ";
 for (int k = 0; k < 3; ++k) {
    //   cerr << _ekfprobtmp[k] << " ";
    _ekfprob[k] = _ekfprobtmp[k];  
    if (_ekfprob[k] < EP)
       _ekfprob[k] = EP;
 }
// cerr << endl;
}


Vec3d
MMAEMultiParam::predictPos(double pinterval)
{
 Vec3d ppos(0.0,0.0,0.0);
 if (_cmb == tMAX) {
    // find the highest probability and use the coresponding state vector
    // and kf model to predict
    int modnum = getHighestModelKF();    
    if (modnum == 0) { 
       // make predicton
       MMatrix pmat = _kfmod[0].getFundMat();
       pmat(0,1) = pmat(2,3) = pmat(4,5) = pinterval;
       MMatrix tmp = pmat * _kfmod[0].getStateVector();
       for(int i = 0; i < 3; ++i)
	  ppos[i] = tmp(i*2,0); 	
      // restart other filters
      if (_kfprobtmp[1] < EP) {
	 for (int j = 0; j < 3; ++j) {
           _kfmod[1].getStateVector()(j*2,0) = _kfmod[0].getStateVector()(j*2,0);
           _kfmod[1].getStateVector()((j*2)+1,0) = _kfmod[0].getStateVector()((j*2)+1,0);
           _kfmod[1].getCov()(j*2,j*2) = _kfmod[0].getCov()(j*2,j*2);
	   _kfmod[1].getCov()((j*2)+1,(j*2)+1) = _kfmod[0].getCov()((j*2)+1,(j*2)+1);
	 }
      }  
      if (_kfprobtmp[2] < EP) {
         for (int j = 0; j < 3; ++j) {
          _kfmod[2].getStateVector()(j*2,0) = _kfmod[0].getStateVector()(j*2,0);
          _kfmod[2].getStateVector()((j*2)+1,0) = _kfmod[0].getStateVector()((j*2)+1,0); 
          _kfmod[2].getCov()(j*2,j*2) = _kfmod[0].getCov()(j*2,j*2);
          _kfmod[2].getCov()((j*2)+1,(j*2)+1) = _kfmod[0].getCov()((j*2)+1,(j*2)+1);
         }
      }
    }
    else if (modnum == 1) {
      // make prediction
      MMatrix pmat = _kfmod[1].getFundMat();
      pmat(0,1) = pmat(2,3) = pmat(4,5) = pinterval;
      MMatrix tmp = pmat * _kfmod[1].getStateVector();
      for(int i = 0; i < 3; ++i)
         ppos[i] = tmp(i*2,0); 
      // restart other filters
      if (_kfprobtmp[0] < EP) {
	 for (int j = 0; j < 3; ++j) {
	   _kfmod[0].getStateVector()(j*2,0) = _kfmod[1].getStateVector()(j*2,0);
           _kfmod[0].getStateVector()((j*2)+1,0) = _kfmod[1].getStateVector()((j*2)+1,0); 
           _kfmod[0].getCov()(j*2,j*2) = _kfmod[1].getCov()(j*2,j*2);
           _kfmod[0].getCov()((j*2)+1,(j*2)+1) = _kfmod[1].getCov()((j*2)+1,(j*2)+1);
         }
       }
       if (_kfprobtmp[2] < EP) {
	 for (int j = 0; j < 3; ++j) {
	   _kfmod[2].getStateVector()(j*2,0) = _kfmod[1].getStateVector()(j*2,0);
           _kfmod[2].getStateVector()((j*2)+1,0) = _kfmod[1].getStateVector()((j*2)+1,0);
           _kfmod[2].getCov()(j*2,j*2) = _kfmod[1].getCov()(j*2,j*2);
           _kfmod[2].getCov()((j*2)+1,(j*2)+1) = _kfmod[1].getCov()((j*2)+1,(j*2)+1);
          }
       }
    }
    else {
       // make prediction
       MMatrix pmat = _kfmod[1].getFundMat();
       pmat(0,1) = pmat(2,3) = pmat(4,5) = pinterval;
       MMatrix tmp = pmat * _kfmod[2].getStateVector();
       for(int i = 0; i < 3; ++i)
         ppos[i] = tmp(i*2,0); 
      // restart other filters
      if (_kfprobtmp[0] < EP) {
         for (int j = 0; j < 3; ++j) {
          _kfmod[0].getStateVector()(j*2,0) = _kfmod[2].getStateVector()(j*2,0);
          _kfmod[0].getStateVector()((j*2)+1,0) = _kfmod[2].getStateVector()((j*2)+1,0); 
          _kfmod[0].getCov()(j*2,j*2) = _kfmod[2].getCov()(j*2,j*2);
          _kfmod[0].getCov()((j*2)+1,(j*2)+1) = _kfmod[2].getCov()((j*2)+1,(j*2)+1);
         }
      }
      if (_kfprobtmp[1] < EP) {
	for (int j = 0; j < 3; ++j) {
	  _kfmod[1].getStateVector()(j*2,0) = _kfmod[2].getStateVector()(j*2,0); 
          _kfmod[1].getStateVector()((j*2)+1,0) = _kfmod[2].getStateVector()((j*2)+1,0); 
          _kfmod[1].getCov()(j*2,j*2) = _kfmod[2].getCov()(j*2,j*2);
          _kfmod[1].getCov()((j*2)+1,(j*2)+1) = _kfmod[2].getCov()((j*2)+1,(j*2)+1);
	}
      }
    }
 }
 else if (_cmb == tSUM) {
    // there a number of ways to do this
    // add up the position

    Vec3d pp1(0.0,0.0,0.0);
    Vec3d pp2(0.0,0.0,0.0);
    Vec3d pp3(0.0,0.0,0.0);
    
    // add up velocities
    Vec3d v1(0.0,0.0,0.0);
    Vec3d v2(0.0,0.0,0.0);
    Vec3d v3(0.0,0.0,0.0);

    for (int i = 0; i < 3; ++i) {
      pp1[i] = _kfmod[0].getStateVector()(i*2,0);
      pp2[i] = _kfmod[1].getStateVector()(i*2,0);
      pp3[i] = _kfmod[2].getStateVector()(i*2,0);
      v1[i]  = _kfmod[0].getStateVector()((i*2)+1,0);
      v2[i]  = _kfmod[1].getStateVector()((i*2)+1,0); 
      v3[i]  = _kfmod[2].getStateVector()((i*2)+1,0);
    }
     
    // blend
    Vec3d pp4 = _kfprobtmp[0] * pp1 + _kfprobtmp[1] * pp2 + _kfprobtmp[2] * pp3;
    Vec3d v4 = _kfprobtmp[0] * v1 + _kfprobtmp[1] * v2 + _kfprobtmp[2] * v3;

    for (int i = 0; i < 3; ++i) {
      if (_kfprobtmp[0] < EP) {
	 _kfmod[0].getStateVector()(i*2,0) = pp4[i];
         _kfmod[0].getStateVector()((i*2)+1,0) = v4[i];
      }
      if (_kfprobtmp[1] < EP) {
	 _kfmod[1].getStateVector()(i*2,0) = pp4[i];
         _kfmod[1].getStateVector()((i*2)+1,0) = v4[i];
      }
      if (_kfprobtmp[2] < EP) {
	 _kfmod[2].getStateVector()(i*2,0) = pp4[i];
	 _kfmod[2].getStateVector()((i*2)+1,0) = v4[i];
      }
    }
    
    MMatrix nstate(6,1);
    for (int i = 0; i < 3; ++i) {
      nstate(i*2,0) = pp4[i];
      nstate((i*2)+1, 0) = v4[i];
    }

    MMatrix pmat = _kfmod[1].getFundMat();
    pmat(0,1) = pmat(2,3) = pmat(4,5) = pinterval;
    MMatrix tmp = pmat * nstate;
    for(int i = 0; i < 3; ++i)
       pp2[i] = tmp(i*2,0); 
    
    for (int i = 0; i < 3; ++i)
      ppos[i] = pp2[i]; 
 }
 return ppos;
}

Quat
MMAEMultiParam::predictQuat(double pinterval)
{
 Quat prot(0.0,0.0,0.0,1.0);

  if (_cmb == tMAX) {
    // find the highest probability and use the coresponding state vector
    // and ekf model to predict
    int modnum = getHighestModelEKF();    
    if (modnum == 0) {
       // make predicton	
       MMatrix estate = _ekfmod[0].getStateVector();
       Quat tq(estate(0,0), estate(1,0), estate(2,0), estate(3,0));
       Quat tw(estate(4,0), estate(5,0), estate(6,0), 0);
       prot = tq + (pinterval * 0.5 * (tq * tw));
      // restart other filters
      if (_ekfprobtmp[1] < EP) {
	 for (int j = 0; j < 4; ++j) {
	   _ekfmod[1].getStateVector()(j,0) = _ekfmod[0].getStateVector()(j,0);
	   //  _ekfmod[1].getCov()(j,j) = _ekfmod[0].getCov()(j,j);
	 }
      }  
      if (_ekfprobtmp[2] < EP) {
         for (int j = 0; j < 4; ++j) {
           _ekfmod[2].getStateVector()(j,0) = _ekfmod[0].getStateVector()(j,0);
	   //  _ekfmod[2].getCov()(j,j) = _ekfmod[0].getCov()(j,j);
         }
      }
    }
    else if (modnum == 1) {
       // make prediction
        MMatrix estate = _ekfmod[1].getStateVector();
        Quat tq(estate(0,0), estate(1,0), estate(2,0), estate(3,0));
        Quat tw(estate(4,0), estate(5,0), estate(6,0), 0);
        prot = tq + (pinterval * 0.5 * (tq * tw));
       // restart other filters
      if (_ekfprobtmp[0] < EP) {
	  for (int j = 0; j < 4; ++j) {
            _ekfmod[0].getStateVector()(j,0) = _ekfmod[1].getStateVector()(j,0);
	    //  _ekfmod[0].getCov()(j,j) = _ekfmod[1].getCov()(j,j);
          }
       }
       if (_ekfprobtmp[2] < EP) {
	 for (int j = 0; j < 7; ++j) {
	   _ekfmod[2].getStateVector()(j,0) = _ekfmod[1].getStateVector()(j,0);
	   // _ekfmod[2].getCov()(j,j) = _ekfmod[1].getCov()(j,j);
         }
       }
    }
    else {  // PVASM has highest probability
      // make prediction
      MMatrix estate = _ekfmod[2].getStateVector();
      Quat tq(estate(0,0), estate(1,0), estate(2,0), estate(3,0));
      Quat tw(estate(4,0), estate(5,0), estate(6,0), 0);
      prot = tq + (pinterval * 0.5 * (tq * tw));

      // restart other filters
      if (_ekfprobtmp[0] < EP) {
         for (int j = 0; j < 4; ++j) {
	   _ekfmod[0].getStateVector()(j,0) = _ekfmod[2].getStateVector()(j,0);
	   // _ekfmod[0].getCov()(j,j) = _ekfmod[2].getCov()(j,j);
         }
      }
      if (_ekfprobtmp[1] < EP) {
	for (int j = 0; j < 7; ++j) {
	  _ekfmod[1].getStateVector()(j,0) = _ekfmod[2].getStateVector()(j,0); 
	  //  _ekfmod[1].getCov()(j,j) = _ekfmod[2].getCov()(j,j);
	}
      }
    }
  }
  else if (_cmb == tSUM) {
    // there a number of ways to do this
    // compute weighted sums of quats and ang vel's

    Quat q1(0.0,0.0,0.0,1.0);
    Quat q2(0.0,0.0,0.0,1.0);
    Quat q3(0.0,0.0,0.0,1.0);
    
    // add up velocities
    Quat av1(0.0,0.0,0.0,0.0);
    Quat av2(0.0,0.0,0.0,0.0);
    Quat av3(0.0,0.0,0.0,0.0);

    for (int i = 0; i < 3; ++i) {
      q1.V()[i] = _ekfmod[0].getStateVector()(i,0);
      q2.V()[i] = _ekfmod[1].getStateVector()(i,0);
      q3.V()[i] = _ekfmod[2].getStateVector()(i,0);
      av1.V()[i]  = _ekfmod[0].getStateVector()(i+4,0);
      av2.V()[i]  = _ekfmod[1].getStateVector()(i+4,0); 
      av3.V()[i]  = _ekfmod[2].getStateVector()(i+4,0);
    }
    q1.S() = _ekfmod[0].getStateVector()(3,0);
    q2.S() = _ekfmod[1].getStateVector()(3,0);
    q3.S() = _ekfmod[2].getStateVector()(3,0);
     
    // blend
    Quat q4 = _ekfprobtmp[0] * q1 + _ekfprobtmp[1] * q2 + _ekfprobtmp[2] * q3;
    Quat av4 = _ekfprobtmp[0] * av1 + _ekfprobtmp[1] * av2 + _ekfprobtmp[2] * av3;

    for (int i = 0; i < 3; ++i) {
       if (_ekfprobtmp[0] < EP) {
	 _ekfmod[0].getStateVector()(i,0) = q4.V()[i];
         if (i == 0)
           _ekfmod[0].getStateVector()(3,0) = q4.S();
       }
      if (_ekfprobtmp[1] < EP) {
	 _ekfmod[1].getStateVector()(i,0) = q4.V()[i];
         if (i == 0)
           _ekfmod[1].getStateVector()(3,0) = q4.S();
         _ekfmod[1].getStateVector()(i+4,0) = av4.V()[i];
      }
      if (_ekfprobtmp[2] < EP) {
	 _ekfmod[2].getStateVector()(i,0) = q4.V()[i];
         if (i == 0)
           _ekfmod[2].getStateVector()(3,0) = q4.S();
	 _ekfmod[2].getStateVector()(i+4,0) = av4.V()[i];
      }
    }
    
    MMatrix estate(7,1);
    for (int i = 0; i < 3; ++i) {
      estate(i,0) = q4.V()[i];
      estate(i+4, 0) = av4.V()[i];
    }
    estate(3,0) = q4.S();

    Quat tq(estate(0,0), estate(1,0), estate(2,0), estate(3,0));
    Quat tw(estate(4,0), estate(5,0), estate(6,0), 0);
    prot = tq + (pinterval * 0.5 * (tq * tw));
  }  
  return prot; 
}


Pose * 
MMAEMultiParam::filter(Pose *userpos, double pinterval)
{
 if (_ctr <= MAXBEGIN-LESSNUM)
   ++_ctr;

 if (_ctr == 1) { // initialize state vectors 
    for (int i = 0; i < 3; ++i) 
       for (int j = 0; j < 3; ++j) {
	 _kfmod[i].getStateVector()(j*2,0) = userpos->_pos[j];
	 _ekfmod[i].getStateVector()(j,0) = userpos->_orient.V()[j];
       }
    for (int i = 0; i < 3; ++i)
       _ekfmod[i].getStateVector()(3,0) = userpos->_orient.S();
    return userpos;
 }  
 
 // run each filter 
 for (int i = 0; i < 3; ++i) {
    _kfmod[i].mmaeKalmanFilter(userpos, 3);
    _ekfmod[i].mmaeExtendedKalmanFilter(userpos,4);
 }

  // compute the conditional probabilities
 computeCondProbKF();
 computeCondProbEKF();
   
 if (_ctr < MAXBEGIN-LESSNUM) // let it run for a little bit before we predict
   return userpos;


 // predict and reinitialize if needed
 Vec3d ppos = predictPos(pinterval);
 Quat prot = predictQuat(pinterval);   
    
 return new Pose(userpos->_tstamp + pinterval, ppos, normalize(prot)); 
}
