// This header contains descriptions of all the UI elements in the
// program.

#ifndef UI_H
#define UI_H

#include "types.H"
#include <qwidget.h>
#include <qmainwindow.h>
#include <qlcdnumber.h>
#include <qslider.h>
#include <qlabel.h>
#include "engine.h"

#include "matrix.h"

class QMultiLineEdit;
class QLabel;
class QMenuBar;
class QStatuBar;
class QString; 
class QLCDNumber;
class QPopupMenu;
class QPushButton;
class QProgressDialog;

class ControlPanel; 
class PlotsDialog;
class LocalPolyDialog;
class ExpSmoothDialog;
class GreyPredDialog;
class KalmanPredDialog;
class CIPredDialog;
class MMAEPredDialog;
class SignalAnalysisDialog;
class MATLABEngine;
class UMotionRec;
class Dispenser;
class SignalAnalyzer;
class NumDiff;
class NDiffToolDialog;

// make special QLCDNUMBER Class 
class QLCDNumberD : public QLCDNumber
{
  Q_OBJECT
 public:
 
   QLCDNumberD(QWidget *parent = 0, const char *name=0) : QLCDNumber(parent,name)
    { }
   virtual ~QLCDNumberD() { }

   public slots:
    void myDisplay(int num);
    void myDisplay2(int num);
    void myDisplay3(int num);
};


// this is the main widget class that holds all widgets and
// can spawn other widget based on events

class GlobalWidget : public QWidget
{
    Q_OBJECT
  public: 
    GlobalWidget(QWidget *parent = 0, const char *name=0);
    virtual ~GlobalWidget();

  protected slots:  
  // file slots
   void loadFile();
   void saveLog();
   void myquit();
  // predictor slots
   void reflection();
   void localPolynomial();
   void expSmoothing();
   void greyPred();
   void kalmanPred();
   void ciPred();
   void mmaePred();
  // error metric slots
   void rootMSE();
   void lInfinity();
   void pctBetter();
   void chooseAll();
  // testing slots
   void singleRun();
   void runParams();
   void runSampleRate();
   void runPredictInt();
   void runNoise();
   void runFull();
  // matlab slots
   void startMatLab();
   void showPlot();
   void spTool();
  // signal analysis slots
   void analyzers(); // starts a dialog for running different analyzers
   void nDiffTools(); // start the num diff tool dialog

  // non slot functions  
   protected:
    void loadMotionData(const char *file, int length);
    void saveLogFile(QString filename);
    void setPredictor();
    void sendUserDataToMatlab();    

  // accessors
  public:
    MATLABEngine     * getMEngine()     { return _mengine; }
    UMotionRec       * getUserRec()     { return _urec; }
    int                getDataPresent() { return _datapresent; }
    QMultiLineEdit   * getTextBox()     { return _textbox; }
    QStatusBar       * getStatusBar()   { return _status; }
    QProgressDialog  * getProgDialog()     { return _processd; }
    PredictorType      getPredType()    { return _ptype; }
    ErrorType          getErrorType()   { return _etype; }
    TestType           getTestType()    { return _rtype; }
    ControlPanel     * getCPanel()      { return _cpanel; }
    PlotsDialog      * getPlotsDialog() { return _plotsd; }
    LocalPolyDialog  * getLPolyDialog() { return _lpolyd; }
    ExpSmoothDialog  * getExpSmDialog() { return _expd; }
    GreyPredDialog   * getGreyPredDialog()  { return _greyd; }
    KalmanPredDialog * getKalmanPredDialog() { return _kpredd; }
    CIPredDialog     * getCIPredDialog() { return _cipredd; }
    MMAEPredDialog   * getMMAEPredDialog() { return _mmaepredd; }
    Dispenser        * getDispenser()   { return _disp; }
    SignalAnalysisDialog * getSigAnD() { return _sigad; }
    NDiffToolDialog  * getNDiffToolDialog() { return _ndiffd; }

  protected:  
   QProgressDialog *_processd;
   QMenuBar       *_menu;
   QStatusBar     *_status;
   QMultiLineEdit *_textbox;
   QPopupMenu     *_filemenu;
   QPopupMenu     *_predictor;
   QPopupMenu     *_errormet;
   QPopupMenu     *_testing;
   QPopupMenu     *_mlab;
   QPopupMenu     *_siga;
   ControlPanel   *_cpanel;
   MATLABEngine   *_mengine; // holds the matlab engine
   UMotionRec     *_urec;    // holds the current motion record
   PredictorType   _ptype;   // predictor type
   ErrorType       _etype;   // error type
   TestType        _rtype;   // testing type
   int             _matlabon;
   int             _dataready;
   int             _datapresent;

   // dialog objects
   PlotsDialog      *_plotsd; 
   LocalPolyDialog  *_lpolyd;
   ExpSmoothDialog  *_expd;
   GreyPredDialog   *_greyd;
   KalmanPredDialog *_kpredd;
   CIPredDialog     *_cipredd;
   MMAEPredDialog   *_mmaepredd; 
   SignalAnalysisDialog *_sigad;
   NDiffToolDialog *_ndiffd;
 
   // the dispenser object
   Dispenser      *_disp;
};


// this class defines a group of widgets that make up at status and control
// panel for global params, it is attached to the main window

class ControlPanel : public QWidget
{
   Q_OBJECT
 public:
   ControlPanel(QWidget *parent = 0, const char *name = 0);
   virtual ~ControlPanel();

  // accessors
  QLabel * getCurFile()      { return _curfile; }
  QLabel * getOrigSP()       { return _origsp; }
  QLabel * getCurSP()        { return _cursp; }
  QLabel * getCurPredictor() { return _curpred; } 
  QLabel * getCurPredInt()   { return _curpredint; }
  QLabel * getCurError()     { return _curerr; }
  QLCDNumber  * getSPLCD()    { return _splcd; } 
  QLCDNumber  * getLeadLCD()  { return _leadt; }
  QLCDNumberD * getNoiseLCD() { return _noisev; }  

  double   getNoiseVar()     { return _noisevar; }   
  double   getCurSRate()     { return _cursrate; }  
  double   getOrigSRate()    { return _origsrate; }  
  double   getLeadTime()     { return _leadtime; }
  int      getSRateInt()     { return _srateint; }
  int      getLeadTimeInt()  { return _leadtimeint; }
   
  void     setOrigSRate(double d) { _origsrate = d; }

 public slots:
   void updateNoise(int num);
   void updateCurSRate(int num);
   void updateLeadTime(int num);
   void updateNoiseD(double num);

 protected:
  QLabel  *_filelabel;
  QLabel  *_curfile;
  QLabel  *_predlabel;
  QLabel  *_curpred;
  QLabel  *_sprate;
  QLabel  *_origsp;
  QLabel  *_newsprate;
  QLabel  *_cursp;
  QLabel  *_predint;
  QLabel  *_curpredint;
  QLabel  *_errormet;
  QLabel  *_curerr;
  
  QLCDNumber  *_splcd;
  QLCDNumber  *_leadt;
  QLCDNumberD *_noisev; 

  double   _noisevar;
  double   _cursrate;
  double   _origsrate;
  double   _leadtime;
  int      _srateint;
  int      _leadtimeint;
};


// a class for handling the matlab plots dialog
class PlotsDialog : public QWidget
{
  Q_OBJECT
 public:
   PlotsDialog(QWidget *parent = 0, const char *name = 0);
   virtual ~PlotsDialog() { }
 
 public slots:
   void showDialog() { show(); }
   void hideDialog() { hide(); }
   void toggleX(bool on) { _showx = on; }
   void toggleY(bool on) { _showy = on; }
   void toggleZ(bool on) { _showz = on; }
   void toggleQX(bool on) { _showqx = on; }
   void toggleQY(bool on) { _showqy = on; }
   void toggleQZ(bool on) { _showqz = on; } 
   void toggleQW(bool on) { _showqw = on; } 
 
 public:  // acessors
   int getX() { return _showx; }  
   int getY() { return _showy; }
   int getZ() { return _showz; }
   int getQX() { return _showqx; }  
   int getQY() { return _showqy; }
   int getQZ() { return _showqz; }
   int getQW() { return _showqw; }
   int plot()  { if (_showx || _showy || _showz || _showqx ||
                     _showqy || _showqz || _showqw)
                   return 1;
                 else 
                   return 0;
               }


 protected:
  QPushButton *_closebtn;
  int _showx;  
  int _showy;
  int _showz;
  int _showqx;
  int _showqy;
  int _showqz;
  int _showqw;

};

// class for handling the local poly predictor dialog
class LocalPolyDialog : public QWidget
{
 Q_OBJECT
  public:
   LocalPolyDialog(QWidget *parent = 0, const char *name = 0);
   virtual ~LocalPolyDialog() { }

   int getPDegree() { if (_polytype == tlagrangian)
                       return _pwindow; 
                      else if (_polytype == tlsquares)
                       return _pdegree;
                      else if (_polytype == tcspline)
                       return 3; 
                      else
                       return 2;
                    }
  int getPWindow() { if (_polytype == ttaylor) 
                       if (_derfly) // need a window for derivatives
                         return 5;
                       else 
                         return 1;// taylor expansion has a window of one
                     else if (_polytype == tcspline)
                        if (_derfly)
                          return _pwindow+4;  // need room to compute ders
                        else
			 return _pwindow;
                     else 
                       return _pwindow; 
                    }

   int getDerFly() { return _derfly; }
   int getKFFlag() { return _kf; }
   LocalPolyType getLPType() { return _polytype; }

  public slots:
    void showDialog()  { show(); }
    void hideDialog()  { hide(); }
    void hideDeg() { _deg->hide(); _degslider->hide(); _deglabel->hide(); }
    void showDeg() { _deg->show(); _degslider->show(); _deglabel->show(); } 
                  
    void setLagrange() { _polytype = tlagrangian; }
    void setLS()       { _polytype = tlsquares; }
    void setCSpline()  { _polytype = tcspline; }
    void setTaylor()   { _polytype = ttaylor; }
    void setDegree(int value) { _pdegree = value; }
    void setWindow(int value) { _pwindow = value; }
    void toggleFly(bool on) { _derfly = on; }
    void toggleKF(bool on) { _kf = on; }

  protected:
   LocalPolyType  _polytype;
   QPushButton   *_closebtn;
   QLCDNumber    *_deg;
   QSlider       *_degslider;
   QLabel        *_deglabel;
   QLCDNumber    *_window;
   int            _pdegree;
   int            _pwindow;
   int            _derfly;
   int            _kf;

};


class ExpSmoothDialog : public QWidget
{
 Q_OBJECT
 public:
   ExpSmoothDialog(QWidget *parent = 0, const char *name = 0);
   virtual ~ExpSmoothDialog() { }
 
   double         getAlpha() { return _alpha; }
   double         getBeta()  { return _beta; }
   int            getKFFlag() { return _kf; }
   ExpSmoothType  getExpSmoothType() { return _expsmtype; }

 public slots: 
   void showDialog() { show(); }
   void hideDialog() { hide(); }
 
   void setSimpleExp()    { _expsmtype = tSimpleExp; }
   void setDoubleExp()    { _expsmtype = tDoubleExp; }
   void setHoltWinters()  { _expsmtype = tHoltWinters; }
   void setAlpha(int val);
   void setBeta(int val);   
   void toggleKF(bool on) { _kf = on; }

 protected:
   ExpSmoothType  _expsmtype; 
   double         _alpha;
   double         _beta;
   QPushButton   *_closebtn; 
   int            _kf;
};


class GreyPredDialog : public QWidget
{
 Q_OBJECT
 public:
   GreyPredDialog(QWidget *parent = 0, const char *name = 0);
   virtual ~GreyPredDialog() { }
   
   int getNumPts() { return _numpts; }
   int getKFFlag() { return _kf; }

 public slots:
   void showDialog() { show(); }
   void hideDialog() { hide(); }
   void setNumPts(int value) { _numpts = value; }
   void toggleKF(bool on) { _kf = on; }

 protected:   
   int          _numpts;
   int          _kf;
   QPushButton *_closebtn;
};


class KalmanPredDialog : public QWidget
{
 Q_OBJECT
 public:
   KalmanPredDialog(QWidget *parent = 0, const char *name = 0);
   virtual ~KalmanPredDialog() { }

   int getSpectralDensity() { return _specd; }
   int getUKFFlag()      { return _ukfon; }
   KalmanType getKalmanType() { return _ktype; }
   
 public slots:
   void showDialog() { show(); }
   void hideDialog() { hide(); }

   void setSpecD(int value) { _specd = value; }
   void setPVSM()    { _ktype = tPVSM; }
   void setPVASM()   { _ktype = tPVASM; }
   void setPVMMV()   { _ktype = tPVMMV; }
   void setPVAMMV()  { _ktype = tPVAMMV; }
   void setPVAMMA()  { _ktype = tPVAMMA; }
   void setPVAMMVA() { _ktype = tPVAMMVA; }
   void toggleUKF(bool on) { _ukfon = on; }
   
   
 protected:
  KalmanType   _ktype;
  int          _specd;
  int          _ukfon;
  QPushButton *_closebtn;
};

// Covariance Intersection Dialog
class CIPredDialog : public QWidget
{
 Q_OBJECT
 public:
   CIPredDialog(QWidget *parent = 0, const char *name = 0);
   virtual ~CIPredDialog() { }

   int getCISpectralDensity() { return _specd; }
   CIType getCIType() {return _citype; }

 public slots: 
   void showDialog() { show(); }
   void hideDialog() { hide(); }
  
   void setCISpecD(int value) { _specd = value; }
   void setCIPVSM()    { _citype = tCIPVSM; }
   void setCIPVMMV() { _citype = tCIPVMMV; }   

 protected:
   CIType       _citype;
   int          _specd;
   QPushButton *_closebtn;
};


// Multiple Model Adaptive Estimator Dialog
class MMAEPredDialog : public QWidget
{
 Q_OBJECT
 public:
   MMAEPredDialog(QWidget *parent = 0, const char *name = 0);
   virtual ~MMAEPredDialog() { }

   int getMMAESpectralDensity() { return _specd; }
   int getMMAESpecDSeparator() { return _seperator; }
   MMAEType     getMMAEType() { return _mmaetype; }
   CombinerType getCombinerType() { return _combiner; }      

 public slots:
   void showDialog() { show(); }
   void hideDialog() { hide(); }
   
   void setMMAESpecD(int value) { _specd = value; }
   void setMMAESpecDSep(int value) { _seperator = value; }
   void setMMAE3Mod() { _mmaetype = tMMAE3MOD; }
   void setMMAE3ModComp() { _mmaetype = tMMAE3MODComp; }
   void setMMAEMulParam() { _mmaetype = tMMAEMulParam; }

   void setSUM() { _combiner = tSUM; }
   void setMAX() { _combiner = tMAX; }   
 
 protected: 
   MMAEType     _mmaetype;
   int          _specd;
   int          _seperator;
   CombinerType _combiner; 
   QPushButton *_closebtn;     
};


// a class for any ui components corresponding to matlab functionality

class MATLABEngine
{
  public: 
    MATLABEngine() : _eng(NULL) { }
    virtual ~MATLABEngine() { }
 
    int startEngine();
    void killEngine();
    int runCmd(QString cmd);
    int sendArray(const mxArray *mp);
    mxArray * getArray(const char *name);
    void createOutputBuffer(char *buf, int size); 

    Engine * getEngine() { return _eng; }
 
  protected:
    Engine *_eng;    


};


#include <iostream.h>
// this class will define the ui for controlling signal analysis and quantification
// tools

class SignalAnalysisDialog : public QWidget
{
 Q_OBJECT
 public:
   SignalAnalysisDialog(GlobalWidget *gw, QWidget *parent = 0, const char *name = 0);
   virtual ~SignalAnalysisDialog();

   AnalyzerType getAnalyzer() { return _atype; }
   int getLag() { return _lag; }
   int getInitWin() { return _initwin; }
   SignalAnalyzer * getSigAnalyzer() { return _analyzer; }

 public slots:
    void showDialog() { show(); }
    void hideDialog() { hide(); }
    void exec();  // calls the appropriate routine

   void setLagValue(int value) { _lag = 10*value;}  
   void setCutOff(int val) { _initwin = val; }
   void setSACF() { _atype = tSACF; }
   void setACFWindowD() { _atype = tACFWin; }
   void setBasicSMetrics() { _atype = tBasic; }
   void setPSD() { _atype = tPSD; }
   void setCrPt() { _atype = tCrPt; }
   void setVelBasedMetrics() { _atype = tVelBasic; }
  
 protected:  
    GlobalWidget   *_gbw;
    SignalAnalyzer *_analyzer;
    AnalyzerType    _atype;
    int             _lag;
    int             _initwin;
    QPushButton    *_closebtn;
    QPushButton    *_execbtn; 
    
};   


// this class is used to control the numerical derivative tools
// for analyzing different num diff schemes

class NDiffToolDialog : public QWidget
{
 Q_OBJECT
 public:
   NDiffToolDialog(GlobalWidget *gw, QWidget *parent = 0, 
                   const char *name = 0);
   virtual ~NDiffToolDialog();

   NDiffType getNDiff() { return _ntype; }
   ThreshType getThreshType() { return _threshtype; }
 
   int getWin()   { return _win; }
   int getDeg()   { return _deg; }
   int getLeft()  { return _left; }
   int getRight() { return _right; } 
   int getCDiff() { return _cdiff; }
   int getHSThresh() { return _hs; }
   int getDerAvg() { return _avgdiff; }

   public slots:
    void showDialog() { show(); }
    void hideDialog() { hide(); }
    void setWin(int value)   { _win = value; }
    void setDeg(int value)   { _deg = value; }
    void setLeft(int value)  { _left = value; }
    void setRight(int value) { _right = value; }
    // set algs
    void setSimpleND() {_ntype = tNDiff; }
    void setLSOrtho() { _ntype = tLSOrthog;}
    void setSavitskyGolay() { _ntype = tSavGolay; }
    void setWaveletDB4()  { _ntype = tWavedb4; }
    void setWaveletDB8()  { _ntype = tWavedb8; }
    void setWaveletSYM8() { _ntype = tWavesym8; }
    void setMollify()     { _ntype = tMollify; }
    void setABOdd()       { _ntype = tABodd; }
    void setABNoOdd()     { _ntype = tABnoodd; }   
    void setKPOdd()       { _ntype = tKPodd; }
    void setKPNoOdd()     { _ntype = tKPnoodd; }
    // set threshholds
    void setRigrSure() { _threshtype = tRigsure; }
    void setHeurSure() { _threshtype = tHeursure; }
    void setSqtwoLog() { _threshtype = tsqtwolog; }
    void setMinimax()  { _threshtype = tMinimax; }
    void setKillDet()  { _threshtype = tkillDet; } 
    void setHardThresh() { _hs = 0; }
    void setSoftThresh() { _hs = 1; }
    void exec();  // calls the appropriate routine   
    bool toggleCDiff(bool on) { _cdiff = on; }
    bool toggleDiffAvg(bool on) { _avgdiff = on; }


 protected:
   GlobalWidget  *_gbw;
   NumDiff       *_ndifftools;
   int            _win;
   int            _deg;
   int            _left;
   int            _right;    
   int            _cdiff;  // flag for using central differencing
   int            _hs;  // flag for hard or soft thresholding
   int            _avgdiff; // flag for derivative averaging
   NDiffType      _ntype;
   ThreshType     _threshtype;
   QPushButton    *_closebtn;
   QPushButton    *_execbtn; 
};

#endif
